import asyncio
from random import sample
from typing import Optional

from surv_ai.lib.conversation.conversation import Conversation
from surv_ai.lib.knowledge_store.interfaces import Knowledge
from surv_ai.lib.llm.interfaces import LargeLanguageModelClientInterface
from surv_ai.lib.log import logger
from surv_ai.lib.tools.interfaces import NoMemoriesFoundException, ToolbeltInterface

from .agents.binary import BinaryAgent
from .agents.reasoning import ReasoningAgent
from .interfaces import SurveyInterface, SurveyResponse


class Survey(SurveyInterface):
    def __init__(
        self,
        client: LargeLanguageModelClientInterface,
        toolbelt: ToolbeltInterface,
        n_agents=10,
        max_concurrency=10,
        max_knowledge_per_agent=5,
        base_knowledge: Optional[list[Knowledge]] = None,
    ):
        self.client = client
        self.toolbelt = toolbelt

        self.n_agents = n_agents
        self.max_concurrency = max_concurrency
        self.max_knowledge_per_agent = max_knowledge_per_agent
        self.base_knowledge = base_knowledge

    async def _poll_agent(
        self,
        statement: str,
        summaries: Conversation,
        relevant_articles: list[Knowledge],
        index: int,
    ):
        try:
            reasoning_agent = ReasoningAgent(
                self.client,
                n_knowledge_items_per_prompt=self.max_knowledge_per_agent,
                _hyperparameters={"temperature": 0.6},
            )

            for article in sample(
                relevant_articles,
                min(len(relevant_articles), self.max_knowledge_per_agent),
            ):
                reasoning_agent.teach_knowledge(article)

            if self.base_knowledge:
                for knowledge in self.base_knowledge:
                    reasoning_agent.teach_knowledge(knowledge)

            response = await reasoning_agent.prompt(statement)

            binary_agent = BinaryAgent(
                self.client,
                _hyperparameters={"temperature": 0.2, "max_tokens": 5},
            )
            binary_agent.teach_text(statement, "Assertion")

            response_conversation = Conversation()
            response_conversation.add(response, f"Researcher #{index}", reasoning_agent.color)

            decision = await binary_agent.prompt(response_conversation)
            summaries.add(decision, f"Researcher #{index}", reasoning_agent.color)

            true_in_decision = "true" in decision.lower()
            false_in_decision = "false" in decision.lower()

            if true_in_decision and not false_in_decision:
                return "true"
            elif false_in_decision and not true_in_decision:
                return "false"
            else:
                return "undecided"
        except NoMemoriesFoundException:
            return "undecided"
        except Exception as e:
            logger.log_exception(e)
            return "error"

    async def conduct(self, hypothesis: str):
        results = {"true": 0, "false": 0, "undecided": 0, "error": 0}

        summaries = Conversation()
        agents = 0

        try:
            relevant_articles = await self.toolbelt.inspect(hypothesis, self.base_knowledge or [])
        except NoMemoriesFoundException:
            return SurveyResponse(
                in_favor=0,
                against=0,
                undecided=0,
                error=self.n_agents,
                percent_in_favor=0,
                uncertainty=0,
            )

        while agents < self.n_agents:
            coroutines = []

            error_rate = results["error"] / self.n_agents
            if error_rate > 0.25:
                raise Exception("Agent error rate is unusually high, likely an issue with API access.")

            for index in range(min(self.max_concurrency, self.n_agents - agents)):
                coroutines.append(self._poll_agent(hypothesis, summaries, relevant_articles, index))
                agents += 1

            decisions = await asyncio.gather(*coroutines)

            for decision in decisions:
                results[decision] += 1

        if results["true"] + results["false"] == 0:
            percent_in_favor = 0
            uncertainty = 1
        else:
            percent_in_favor = results["true"] / (results["true"] + results["false"])
            uncertainty = results["undecided"] / (results["true"] + results["false"])

        return SurveyResponse(
            in_favor=results["true"],
            against=results["false"],
            undecided=results["undecided"],
            error=results["error"],
            percent_in_favor=percent_in_favor,
            uncertainty=uncertainty,
        )



from typing import Optional, Protocol

from pydantic import BaseModel


class PromptMessage(BaseModel):
    role: str
    content: str
    name: Optional[str]


class Prompt(BaseModel):
    messages: list[PromptMessage]


class LargeLanguageModelClientInterface(Protocol):
    def __init__(self, *args, **kwargs):
        ...

    async def get_completions(self, prompts: list[Prompt], **_hyperparameters) -> list[str]:
        ...



import asyncio
import re

from aiohttp import ClientSession
from bs4 import BeautifulSoup

from surv_ai.lib.knowledge_store.interfaces import Knowledge
from surv_ai.lib.llm.interfaces import (
    LargeLanguageModelClientInterface,
    Prompt,
    PromptMessage,
)
from surv_ai.lib.log import logger

from .interfaces import QueryToolInterface


class WikipediaTool(QueryToolInterface):
    instruction = """
        `WIKIPEDIA(searchTerms)` - use a no more than three keywords to search Wikipedia for additional information.
    """
    command = r"WIKIPEDIA\((.+)\)"

    _base_url = "https://en.wikipedia.org/w/api.php"

    def __init__(
        self,
        llm_client: LargeLanguageModelClientInterface,
        n_articles=1,
    ):
        self.client = llm_client
        self.n_articles = n_articles

        self._already_searched = dict()

    async def _search(self, session, query: str) -> list[str]:
        params = {
            "action": "query",
            "format": "json",
            "list": "search",
            "srsearch": re.sub(r"[^A-Za-z0-9 ]+", "", query),
        }
        async with session.get(self._base_url, params=params) as response:
            data = await response.json()
            return [result["title"] for result in data["query"]["search"]]

    async def _get_page_text(self, session, page_title: str) -> str:
        if page_title in self._already_searched:
            return self._already_searched[page_title]

        params = {
            "action": "parse",
            "format": "json",
            "page": page_title,
            "prop": "text",
        }

        async with session.get(self._base_url, params=params) as response:
            data = await response.json()

            try:
                html_content = data["parse"]["text"]["*"]

                soup = BeautifulSoup(html_content, "html.parser")

                results = []
                for paragraph in soup.find_all("p"):
                    paragraph_text = paragraph.text.strip()

                    if paragraph_text:
                        results.append(paragraph_text)

                self._already_searched[page_title] = results
            except Exception:
                return ""

            return results

    async def _ingest_page_information(
        self,
        session,
        original_prompt: str,
        page_title: str,
    ):
        paragraphs = await self._get_page_text(session, page_title)

        prompt = Prompt(
            messages=[
                PromptMessage(
                    content=f"""A user has asked you some questions:
                    
                    {original_prompt}

                    All subsequent messages will be paragraphs from a Wikipedia article is entitled "{page_title}."

                    Your job is to extract any useful information that might help someone answer these questions.

                    Remember to include as much data and concrete examples as possible from the article.

                    For each useful piece of information you extract, please state why it relates to the original questions.
                    """,
                    role="system",
                ),
                *[PromptMessage(content=paragraph, role="user", name="Article") for paragraph in paragraphs],
                PromptMessage(
                    role="assistant",
                    content=f"I have read the Wikpedia article entitled {page_title} and some useful information is:",
                ),
            ],
        )
        response = await self.client.get_completions([prompt], **{"temperature": 0.2})

        return response[0]

    async def _ingest_pages(
        self,
        session: ClientSession,
        original_prompt: str,
        page_title: str,
    ):
        logger.log_context(f"......Learning Wikipedia page with title {page_title}......")
        page_summary = await self._ingest_page_information(
            session,
            original_prompt,
            page_title,
        )

        source = f"https://en.wikipedia.org/wiki/{page_title.replace(' ', '_')}"

        return Knowledge(
            text=f"Wikipedia page entitled {page_title}: {page_summary}",
            source=source,
        )

    async def _filter_relevant_pages(
        self,
        original_prompt: str,
        search_results: list[str],
    ):
        prompt = Prompt(
            messages=[
                PromptMessage(
                    content=f"""Your job is to determine which, if any of the following article titles are relevant to a user's prompt.
                                   
                    Original prompt:
                    
                    '{original_prompt}'.

                    The next message will be the the list of articles separated by commas.

                    Please return only the relevant article titles, also separated by commas.
                    """,
                    role="system",
                ),
                PromptMessage(
                    content=", ".join(search_results),
                    role="user",
                ),
            ],
        )
        response = (await self.client.get_completions([prompt]))[0]

        return response

    async def use(
        self,
        original_prompt: str,
        search_query: str,
    ) -> list[Knowledge]:
        async with ClientSession() as session:
            search_results = await self._search(session, search_query)

            relevant_results = (await self._filter_relevant_pages(original_prompt, search_results)).split(", ")

            if len(relevant_results) == 0:
                return []

            return await asyncio.gather(
                *[
                    self._ingest_pages(session, original_prompt, page_title)
                    for page_title in relevant_results[: self.n_articles]
                ]
            )



from typing import Optional, Protocol

from pydantic import BaseModel

from ..llm.interfaces import LargeLanguageModelClientInterface


class Knowledge(BaseModel):
    text: str
    source: Optional[str]


class KnowledgeStoreInterface(Protocol):
    def __init__(self, client: LargeLanguageModelClientInterface, *args, **kwargs):
        ...

    def recall_recent(self, n_knowledge_items=5, **kwargs) -> list[Knowledge]:
        ...

    def add_text(self, input: str, source: Optional[str] = None, **kwargs):
        ...

    def add_knowledge(self, knowledge: Knowledge, **kwargs):
        ...

    @staticmethod
    def knowledge_as_string(knowledge: list[Knowledge]) -> str:
        ...



import asyncio

from .interfaces import DataPoint, ModelInterface, SurveyInterface, SurveyParameter


class Model(ModelInterface):
    def __init__(
        self,
        survey_class: type[SurveyInterface],
        parameters: list[SurveyParameter],
        max_concurrency: int = 1,
    ):
        self.survey_class = survey_class
        self.max_concurrency = max_concurrency
        self.parameters = parameters

    async def build(self, hypothesis: str) -> list[DataPoint]:
        survey_index = 0
        results = []

        while survey_index < len(self.parameters):
            coroutines = []

            for _ in range(self.max_concurrency):
                survey = self.survey_class(**self.parameters[survey_index].kwargs)
                coroutines.append(survey.conduct(hypothesis))
                survey_index += 1

            results += await asyncio.gather(*coroutines)

        return [
            DataPoint(parameter=parameter, response=response) for parameter, response in zip(self.parameters, results)
        ]

    @staticmethod
    def get_plot_variables(data_points: list[DataPoint]):
        return (
            [p.parameter.independent_variable for p in data_points],
            [p.response.percent_in_favor for p in data_points],
        )



from surv_ai.lib.knowledge_store.interfaces import Knowledge

from ..interfaces import ToolInterface


class QueryToolInterface(ToolInterface):
    async def use(self, query: str, *args) -> list[Knowledge]:
        ...
